package sun.awt.image.codec;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferByte;
import java.awt.image.ComponentColorModel;
import java.awt.image.RescaleOp;
import java.awt.Point;
import java.awt.image.SinglePixelPackedSampleModel;
import java.awt.image.SampleModel;
import java.awt.image.ComponentSampleModel;
import java.io.IOException;
import java.awt.image.WritableRaster;
import com.sun.image.codec.jpeg.ImageFormatException;
import com.sun.image.codec.jpeg.JPEGDecodeParam;
import java.awt.image.IndexColorModel;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.color.ColorSpace;
import java.awt.image.ColorModel;
import com.sun.image.codec.jpeg.JPEGEncodeParam;
import java.io.OutputStream;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

public class JPEGImageEncoderImpl implements JPEGImageEncoder
{
    private OutputStream outStream;
    private JPEGParam param;
    private boolean pack;
    private static final Class OutputStreamClass;
    
    public JPEGImageEncoderImpl(OutputStream outStream) {
        this.outStream = null;
        this.param = null;
        this.pack = false;
        if (outStream == null) {
            throw new IllegalArgumentException("OutputStream is null.");
        }
        this.outStream = outStream;
        this.initEncoder(JPEGImageEncoderImpl.OutputStreamClass);
        return;
    }
    
    public JPEGImageEncoderImpl(OutputStream outputStream, JPEGEncodeParam jpegEncodeParam) {
        this(outputStream);
        this.setJPEGEncodeParam(jpegEncodeParam);
    }
    
    public int getDefaultColorId(ColorModel colorModel) {
        int hasAlpha = colorModel.hasAlpha();
        ColorSpace colorSpace = colorModel.getColorSpace();
        ColorSpace instance = null;
        switch (colorSpace.getType()) {
            case 6: {
                return 1;
            }
            case 5: {
                if (!(hasAlpha == 0)) {
                    return 7;
                }
                return 3;
            }
            case 3: {
                if (instance == null) {
                    try {
                        instance = ColorSpace.getInstance(1002);
                    }
                    catch (IllegalArgumentException ex) {}
                }
                if (colorSpace == instance) {
                    return hasAlpha ? 10 : 5;
                }
                return hasAlpha ? 7 : 3;
            }
            case 9: {
                return 4;
            }
            default: {
                return 0;
            }
        }
    }
    
    public synchronized OutputStream getOutputStream() {
        return this.outStream;
    }
    
    public synchronized void setJPEGEncodeParam(JPEGEncodeParam jpegEncodeParam) {
        this.param = new JPEGParam(jpegEncodeParam);
    }
    
    public synchronized JPEGEncodeParam getJPEGEncodeParam() {
        return (JPEGEncodeParam)this.param.clone();
    }
    
    public JPEGEncodeParam getDefaultJPEGEncodeParam(Raster raster, int n) {
        JPEGParam jpegParam = new JPEGParam(n, raster.getNumBands());
        jpegParam.setWidth(raster.getWidth());
        jpegParam.setHeight(raster.getHeight());
        return jpegParam;
    }
    
    public JPEGEncodeParam getDefaultJPEGEncodeParam(BufferedImage bufferedImage) {
        ColorModel colorModel = bufferedImage.getColorModel();
        int defaultColorId = this.getDefaultColorId(colorModel);
        if (!(colorModel instanceof IndexColorModel)) {
            return this.getDefaultJPEGEncodeParam(bufferedImage.getRaster(), defaultColorId);
        }
        JPEGParam jpegParam;
        if (colorModel.hasAlpha()) {
            jpegParam = new JPEGParam(defaultColorId, 4);
        }
        else {
            jpegParam = new JPEGParam(defaultColorId, 3);
        }
        jpegParam.setWidth(bufferedImage.getWidth());
        jpegParam.setHeight(bufferedImage.getHeight());
        return jpegParam;
    }
    
    public JPEGEncodeParam getDefaultJPEGEncodeParam(int n, int n2) {
        return new JPEGParam(n2, n);
    }
    
    public JPEGEncodeParam getDefaultJPEGEncodeParam(JPEGDecodeParam jpegDecodeParam) throws ImageFormatException {
        return new JPEGParam(jpegDecodeParam);
    }
    
    public synchronized void encode(BufferedImage convertToIntDiscrete) throws IOException, ImageFormatException {
        if (this.param == null) {
            this.setJPEGEncodeParam(this.getDefaultJPEGEncodeParam(convertToIntDiscrete));
        }
        if (convertToIntDiscrete.getWidth() != this.param.getWidth() || convertToIntDiscrete.getHeight() != this.param.getHeight()) {
            throw new ImageFormatException("Param block's width/height doesn't match the BufferedImage");
        }
        if (this.param.getEncodedColorID() != this.getDefaultColorId(convertToIntDiscrete.getColorModel())) {
            throw new ImageFormatException("The encoded COLOR_ID doesn't match the BufferedImage");
        }
        WritableRaster raster = convertToIntDiscrete.getRaster();
        ColorModel colorModel = convertToIntDiscrete.getColorModel();
        if ((colorModel instanceof IndexColorModel)) {
            convertToIntDiscrete = ((IndexColorModel)colorModel).convertToIntDiscrete(raster, false);
            raster = convertToIntDiscrete.getRaster();
            colorModel = convertToIntDiscrete.getColorModel();
        }
        this.encode(raster, colorModel);
        return;
    }
    
    public synchronized void encode(BufferedImage bufferedImage, JPEGEncodeParam jpegEncodeParam) throws IOException, ImageFormatException {
        this.setJPEGEncodeParam(jpegEncodeParam);
        this.encode(bufferedImage);
    }
    
    public void encode(Raster raster) throws IOException, ImageFormatException {
        if (this.param == null) {
            this.setJPEGEncodeParam(this.getDefaultJPEGEncodeParam(raster, 0));
        }
        if (raster.getNumBands() != raster.getSampleModel().getNumBands()) {
            throw new ImageFormatException("Raster's number of bands doesn't match the SampleModel");
        }
        if (raster.getWidth() != this.param.getWidth() || raster.getHeight() != this.param.getHeight()) {
            throw new ImageFormatException("Param block's width/height doesn't match the Raster");
        }
        if (this.param.getEncodedColorID() != 0 && this.param.getNumComponents() != raster.getNumBands()) {
            throw new ImageFormatException("Param block's COLOR_ID doesn't match the Raster.");
        }
        this.encode(raster, (ColorModel)null);
        return;
    }
    
    public void encode(Raster raster, JPEGEncodeParam jpegEncodeParam) throws IOException, ImageFormatException {
        this.setJPEGEncodeParam(jpegEncodeParam);
        this.encode(raster);
    }
    
    private boolean useGiven(Raster raster) {
        SampleModel sampleModel = raster.getSampleModel();
        if (sampleModel.getDataType() != 0) {
            return false;
        }
        if (!(sampleModel instanceof ComponentSampleModel)) {
            return false;
        }
        ComponentSampleModel componentSampleModel = (ComponentSampleModel)sampleModel;
        if (componentSampleModel.getPixelStride() != sampleModel.getNumBands()) {
            return false;
        }
        int[] bandOffsets = componentSampleModel.getBandOffsets();
        int i = 0;
        while (i < bandOffsets.length) {
            if (bandOffsets[i] != i) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    private boolean canPack(Raster raster) {
        SampleModel sampleModel = raster.getSampleModel();
        if (sampleModel.getDataType() != 3) {
            return false;
        }
        if (!(sampleModel instanceof SinglePixelPackedSampleModel)) {
            return false;
        }
        SinglePixelPackedSampleModel singlePixelPackedSampleModel = (SinglePixelPackedSampleModel)sampleModel;
        int[] array = { 16711680, 65280, 255, -16777216 };
        int[] bitMasks = singlePixelPackedSampleModel.getBitMasks();
        if (bitMasks.length != 3 && bitMasks.length != 4) {
            return false;
        }
        int i = 0;
        while (i < bitMasks.length) {
            if (bitMasks[i] != array[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    private void encode(Raster raster, ColorModel colorModel) throws IOException, ImageFormatException {
        SampleModel sampleModel = raster.getSampleModel();
        int numBands = sampleModel.getNumBands();
        if (colorModel == null) {
            int i = 0;
            while (i < numBands) {
                if (sampleModel.getSampleSize(i) > 8) {
                    throw new ImageFormatException("JPEG encoder can only accept 8 bit data.");
                }
                ++i;
            }
        }
        int encodedColorID = this.param.getEncodedColorID();
        switch (this.param.getNumComponents()) {
            case 1: {
                if (encodedColorID != 1 && encodedColorID != 0 && this.param.findAPP0() != null) {
                    throw new ImageFormatException("1 band JFIF files imply Y or unknown encoding.\nParam block indicates alternate encoding.");
                }
                break;
            }
            case 3: {
                if (encodedColorID != 3 && this.param.findAPP0() != null) {
                    throw new ImageFormatException("3 band JFIF files imply YCbCr encoding.\nParam block indicates alternate encoding.");
                }
                break;
            }
            case 4: {
                if (encodedColorID != 4 && this.param.findAPP0() != null) {
                    throw new ImageFormatException("4 band JFIF files imply CMYK encoding.\nParam block indicates alternate encoding.");
                }
                break;
            }
        }
        if (!this.param.isImageInfoValid()) {
            this.writeJPEGStream(this.param, colorModel, this.outStream, null, 0, 0);
            return;
        }
        DataBuffer dataBuffer = raster.getDataBuffer();
        boolean n = false;
        boolean n2 = true;
        int[] componentSize = null;
        if (colorModel != null) {
            if (colorModel.hasAlpha() && colorModel.isAlphaPremultiplied()) {
                n = true;
                n2 = false;
            }
            componentSize = colorModel.getComponentSize();
            for (int j = 0; j < numBands; ++j) {
                if (componentSize[j] != 8) {
                    n2 = false;
                }
            }
        }
        this.pack = false;
        byte[] array;
        int n3;
        int n4;
        if (n2 && this.useGiven(raster)) {
            ComponentSampleModel componentSampleModel = (ComponentSampleModel)sampleModel;
            n3 = dataBuffer.getOffset() + componentSampleModel.getOffset(raster.getMinX() - raster.getSampleModelTranslateX(), raster.getMinY() - raster.getSampleModelTranslateY());
            n4 = componentSampleModel.getScanlineStride();
            array = ((DataBufferByte)dataBuffer).getData();
        }
        else if (n2 && this.canPack(raster)) {
            SinglePixelPackedSampleModel singlePixelPackedSampleModel = (SinglePixelPackedSampleModel)sampleModel;
            n3 = dataBuffer.getOffset() + singlePixelPackedSampleModel.getOffset(raster.getMinX() - raster.getSampleModelTranslateX(), raster.getMinY() - raster.getSampleModelTranslateY());
            n4 = singlePixelPackedSampleModel.getScanlineStride();
            array = ((DataBufferInt)dataBuffer).getData();
            this.pack = true;
        }
        else {
            int[] bandOffsets = new int[numBands];
            float[] scaleFactors = new float[numBands];
            for (int k = 0; k < numBands; ++k) {
                bandOffsets[k] = k;
                if (n2 == 0) {
                    if (componentSize[k] != 8) {
                        scaleFactors[k] = 255.0f / (float)((1 << componentSize[k]) - 1);
                    }
                    else {
                        scaleFactors[k] = 1.0f;
                    }
                }
            }
            ComponentSampleModel sm = new ComponentSampleModel(0, raster.getWidth(), raster.getHeight(), numBands, numBands * raster.getWidth(), bandOffsets);
            WritableRaster writableRaster = Raster.createWritableRaster(sm, new Point(raster.getMinX(), raster.getMinY()));
            if (!(n2 == 0)) {
                writableRaster.setRect(raster);
            }
            else {
                new RescaleOp(scaleFactors, new float[numBands], null).filter(raster, writableRaster);
                if (!(n == 0)) {
                    int[] bits = new int[numBands];
                    for (int l = 0; l < numBands; ++l) {
                        bits[l] = 8;
                    }
                    new ComponentColorModel(colorModel.getColorSpace(), bits, true, true, 3, 0).coerceData(writableRaster, false);
                }
            }
            dataBuffer = writableRaster.getDataBuffer();
            array = ((DataBufferByte)dataBuffer).getData();
            n3 = dataBuffer.getOffset() + sm.getOffset(0, 0);
            n4 = sm.getScanlineStride();
        }
        this.verify(n3, n4, dataBuffer.getSize());
        this.writeJPEGStream(this.param, colorModel, this.outStream, array, n3, n4);
        return;
    }
    
    private void verify(int n, int i, int n2) throws ImageFormatException {
        int width = this.param.getWidth();
        int height = this.param.getHeight();
        int j = this.pack ? 1 : this.param.getNumComponents();
        if (width <= 0 || height <= 0 || height > Integer.MAX_VALUE / width) {
            throw new ImageFormatException("Invalid image dimensions");
        }
        if (i < 0 || i > Integer.MAX_VALUE / height || i > n2) {
            throw new ImageFormatException("Invalid scanline stride: " + i);
        }
        int n3 = (height - 1) * i;
        if (j < 0 || j > Integer.MAX_VALUE / width || j > n2 || j * width > i) {
            throw new ImageFormatException("Invalid pixel stride: " + j);
        }
        int n4 = width * j;
        if (n4 > Integer.MAX_VALUE - n3) {
            throw new ImageFormatException("Invalid raster attributes");
        }
        int n5 = n3 + n4;
        if (n < 0 || n > Integer.MAX_VALUE - n5) {
            throw new ImageFormatException("Invalid data offset");
        }
        if (n + n5 > n2) {
            throw new ImageFormatException("Computed buffer size doesn't match DataBuffer");
        }
        return;
    }
    
    private int getNearestColorId(ColorModel colorModel) {
        switch (colorModel.getColorSpace().getType()) {
            case 5: {
                if (colorModel.hasAlpha()) {
                    return 6;
                }
                return 2;
            }
            default: {
                return this.getDefaultColorId(colorModel);
            }
        }
    }
    
    private native void initEncoder(Class p0);
    
    private synchronized native void writeJPEGStream(JPEGEncodeParam p0, ColorModel p1, OutputStream p2, Object p3, int p4, int p5) throws IOException, ImageFormatException;
    
    static {
        OutputStreamClass = OutputStream.class;
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                System.loadLibrary("jpeg");
                return null;
            }
        });
    }
}
